import enum
from typing import Any, ClassVar, List, Optional, TypedDict, Union, overload

from ft4222 import GPIO, SPI, I2CMaster, SPIMaster, SPISlave

class FT2XXDeviceError(Exception):
    def __init__(self, msgnum: int) -> None: ...

class FT4222DeviceError(FT2XXDeviceError):
    def __init__(self, msgnum: int) -> None: ...

class SysClock(enum.IntEnum):
    CLK_24: int
    CLK_48: int
    CLK_60: int
    CLK_80: int

class DeviceDetail(TypedDict):
    index: int
    flags: int
    type: int
    id: int
    location: int
    serial: bytes
    description: bytes
    handle: int

def createDeviceInfoList() -> int: ...
def getDeviceInfoDetail(devnum: int, update: bool) -> DeviceDetail: ...
def openBySerial(serial: Union[str, bytes]) -> FT4222: ...
def openByDescription(desc: Union[str, bytes]) -> FT4222: ...
def openByLocation(locId: int) -> FT4222: ...

class FT4222:
    def __init__(self, handle: int, update: bool) -> None: ...
    @property
    def chipRevision(self) -> str: ...
    @property
    def chipVersion(self) -> int: ...
    @property
    def libVersion(self) -> int: ...
    def setTimeouts(self, read_timeout: int, write_timeout: int) -> None: ...
    def close(self) -> None: ...
    def setClock(self, clk: SysClock) -> None: ...
    def getClock(self) -> SysClock: ...
    def setSuspendOut(self, enable: bool) -> None: ...
    def setWakeUpInterrupt(self, enable: bool) -> None: ...
    def vendorCmdGet(self, req: Union[str, bytes], bytesToRead: int) -> bytes: ...
    def vendorCmdSet(
        self, req: Union[str, bytes], data: Union[int, bytes, bytearray]
    ) -> None: ...
    def gpio_Init(
        self,
        *args: Any,
        gpio0: GPIO.Dir,
        gpio1: GPIO.Dir,
        gpio2: GPIO.Dir,
        gpio3: GPIO.Dir
    ) -> None: ...
    def gpio_Read(self, portNum: GPIO.Port) -> bool: ...
    def gpio_Write(self, portNum: GPIO.Port, value: bool) -> None: ...
    def gpio_SetInputTrigger(
        self, portNum: GPIO.Port, trigger: GPIO.Trigger
    ) -> None: ...
    def gpio_GetTriggerStatus(self, portNum: GPIO.Port) -> GPIO.Trigger: ...
    def gpio_ReadTriggerQueue(
        self, portNum: GPIO.Port, readsize: Optional[int]
    ) -> List[GPIO.Trigger]: ...
    def i2cMaster_Init(self, kbps: int) -> None: ...
    def i2cMaster_Read(self, addr: int, bytesToRead: int) -> bytes: ...
    def i2cMaster_ReadEx(self, addr: int, flag: int, bytesToRead: int) -> bytes: ...
    def i2cMaster_Write(self, addr: int, data: Union[int, bytes, bytearray]) -> int: ...
    def i2cMaster_WriteEx(
        self, addr: int, flag: int, data: Union[int, bytes, bytearray]
    ) -> int: ...
    def i2cMaster_Reset(self) -> None: ...
    def i2cMaster_GetStatus(self) -> I2CMaster.ControllerStatus: ...
    def spi_Reset(self) -> None: ...
    def spi_ResetTransaction(self, spiIdx: int) -> None: ...
    def spi_SetDrivingStrength(
        self,
        clkStrength: SPI.DrivingStrength,
        ioStrength: SPI.DrivingStrength,
        ssoStrength: SPI.DrivingStrength,
    ) -> None: ...
    def spiMaster_Init(
        self,
        mode: SPIMaster.Mode,
        clock: SPIMaster.Clock,
        cpol: SPI.Cpol,
        cpha: SPI.Cpha,
        ssoMap: SPIMaster.SlaveSelect,
    ) -> None: ...
    def spiMaster_SetLines(self, mode: SPIMaster.Mode) -> None: ...
    def spiMaster_SingleRead(
        self, bytesToRead: int, isEndTransaction: bool
    ) -> bytes: ...
    def spiMaster_SingleWrite(
        self, data: Union[int, bytes, bytearray], isEndTransaction: bool
    ) -> int: ...
    def spiMaster_SingleReadWrite(
        self, data: Union[int, bytes, bytearray], isEndTransaction: bool
    ) -> bytes: ...
    def spiMaster_MultiReadWrite(
        self,
        singleWrite: Union[int, bytes, bytearray],
        multiWrite: Union[int, bytes, bytearray],
        bytesToRead: int,
    ) -> bytes: ...
    def spiMaster_EndTransaction(self) -> None: ...
    def spiSlave_Init(self) -> None: ...
    def spiSlave_InitEx(self, proto: SPISlave.Protocol) -> None: ...
    def spiSlave_Read(self, bytesToRead: int) -> bytes: ...
    def spiSlave_SetMode(self, cpol: SPI.Cpol, cpha: SPI.Cpha) -> None: ...
    def spiSlave_GetRxStatus(self) -> int: ...
    def spiSlave_Write(self, data: Union[int, bytes, bytearray]) -> int: ...



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#
#cython: language_level=3

cdef extern from "ftd2xx.h":
    ctypedef unsigned int DWORD
    ctypedef unsigned int ULONG
    ctypedef unsigned short USHORT
    ctypedef unsigned short SHORT
    ctypedef unsigned char UCHAR
    ctypedef unsigned short WORD
    ctypedef unsigned char BYTE
    ctypedef BYTE* LPBYTE
    ctypedef unsigned int BOOL
    ctypedef unsigned char BOOLEAN
    ctypedef unsigned char CHAR
    ctypedef BOOL* LPBOOL
    ctypedef UCHAR* PUCHAR
    ctypedef const char* LPCSTR
    ctypedef char* PCHAR
    ctypedef void* PVOID
    ctypedef void* HANDLE
    ctypedef unsigned int LONG
    ctypedef int INT
    ctypedef unsigned int UINT
    ctypedef char* LPSTR
    ctypedef char* LPTSTR
    ctypedef const char* LPCTSTR
    ctypedef DWORD* LPDWORD
    ctypedef WORD* LPWORD
    ctypedef ULONG* PULONG
    ctypedef LONG* LPLONG
    ctypedef PVOID LPVOID
    ctypedef void VOID
    ctypedef unsigned long long int ULONGLONG

    cdef enum:
        FT_OK = 0
        FT_INVALID_HANDLE = 1
        FT_DEVICE_NOT_FOUND = 2
        FT_DEVICE_NOT_OPENED = 3
        FT_IO_ERROR = 4
        FT_INSUFFICIENT_RESOURCES = 5
        FT_INVALID_PARAMETER = 6
        FT_INVALID_BAUD_RATE = 7

        FT_DEVICE_NOT_OPENED_FOR_ERASE = 8
        FT_DEVICE_NOT_OPENED_FOR_WRITE = 9
        FT_FAILED_TO_WRITE_DEVICE = 10
        FT_EEPROM_READ_FAILED = 11
        FT_EEPROM_WRITE_FAILED = 12
        FT_EEPROM_ERASE_FAILED = 13
        FT_EEPROM_NOT_PRESENT = 14
        FT_EEPROM_NOT_PROGRAMMED = 15
        FT_INVALID_ARGS = 16
        FT_NOT_SUPPORTED = 17
        FT_OTHER_ERROR = 18
        FT_DEVICE_LIST_NOT_READY = 19

    cdef enum:
        FT_OPEN_BY_SERIAL_NUMBER = 1
        FT_OPEN_BY_DESCRIPTION = 2
        FT_OPEN_BY_LOCATION = 4

    ctypedef PVOID FT_HANDLE
    ctypedef ULONG FT_STATUS

    FT_STATUS FT_CreateDeviceInfoList(LPDWORD lpdwNumDevs)

    FT_STATUS FT_GetDeviceInfoDetail(DWORD dwIndex, LPDWORD lpdwFlags,
        LPDWORD lpdwType, LPDWORD lpdwID, LPDWORD lpdwLocId, LPVOID lpSerialNumber,
        LPVOID lpDescription, FT_HANDLE *pftHandle)

    FT_STATUS FT_OpenEx(PVOID pArg1, DWORD Flags, FT_HANDLE *pHandle)
    FT_STATUS FT_Close(FT_HANDLE ftHandle)

    FT_STATUS FT_Write(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD dwBytesToWrite, LPDWORD lpBytesWritten);
    FT_STATUS FT_VendorCmdGet(FT_HANDLE ftHandle, UCHAR Request, UCHAR *Buf, USHORT Len);
    FT_STATUS FT_VendorCmdSet(FT_HANDLE ftHandle, UCHAR Request, UCHAR *Buf, USHORT Len);

    FT_STATUS FT_SetTimeouts(FT_HANDLE ftHandle, ULONG ReadTimeout, ULONG WriteTimeout);



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""SPI Slave

Definitions to configure the SPI Slave interface.
"""

from enum import IntEnum

class Protocol(IntEnum):
    """SPI SLAVE protocol

    Attributes:
        SPI_SLAVE_WITH_PROTOCOL: With the full SPI SLAVE PROTOCOL supported
        SPI_SLAVE_NO_PROTOCOL: Remove SPI SLAVE protocol, users can design their own protocol
        SPI_SLAVE_NO_ACK: Retain SPI SLAVE protocol but remove command ‘ACK’

    """
    SPI_SLAVE_WITH_PROTOCOL = 0
    SPI_SLAVE_NO_PROTOCOL   = 1
    SPI_SLAVE_NO_ACK        = 2



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""GPIO

Definitions to configure the GPIO interface.
"""

from enum import IntEnum, IntFlag

class Trigger(IntFlag):
    """GPIO Trigger in 'interrupt' mode

    Attributes:
        RISING: Rising edge
        FALLING: Falling edge
        LEVEL_HIGH: High level
        LEVEL_LOW: Low level

    """
    RISING         = 0x01
    FALLING        = 0x02
    LEVEL_HIGH     = 0x04
    LEVEL_LOW      = 0X08

class Output(IntEnum):
    """GPIO Output

    Attributes:
        LOW: Logic low, 0
        HIGH: Logic high, 1

    """
    LOW  = 0
    HIGH = 1

class Port(IntEnum):
    """GPIO Port

    Attributes:
        P0: Port 0
        P1: Port 1
        P2: Port 2
        P3: Port 3

    """
    P0  = 0
    P1  = 1
    P2  = 2
    P3  = 3

class Dir(IntEnum):
    """GPIO Output direction

    Attributes:
        OUTPUT: Use as output
        INPUT: Use as input

    """
    OUTPUT = 0
    INPUT  = 1



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""I2C Master

Definitions to configure the I2C Master interface.
"""

from enum import IntEnum, IntFlag

class Flag(IntEnum):
    """I2CMaster Flags

    This flags control the start and stopbit generation during a I2C transfer

    Attributes:
        NONE: No start nor stopbit
        START: Startbit
        REPEATED_START: Repeated startbit (will not send master code in HS mode)
        STOP: Stopbit
        START_AND_STOP: Startbit and stopbit

    """
    NONE = 0x80
    START = 0x02
    REPEATED_START = 0x03  # Repeated_START will not send master code in HS mode
    STOP  = 0x04
    START_AND_STOP = 0x06  # START condition followed by SEND and STOP condition

class ControllerStatus(IntFlag):
    """I2CMaster controller Status

    Attributes:
        BUSY:  controller busy: all other status bits invalid
        ERROR: error condition
        ADDRESS_NACK: slave address was not acknowledged during last operation
        DATA_NACK: data not acknowledged during last operation
        ARB_LOST: arbitration lost during last operation
        IDLE: controller idle
        BUS_BUSY: bus busy

    """
    BUSY = 0x01
    ERROR = 0x02
    ADDRESS_NACK = 0x04
    DATA_NACK = 0x08
    ARB_LOST = 0x10
    IDLE = 0x20
    BUS_BUSY = 0x40



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""SPI

Definitions to configure either the SPI Master or Slave interface.
"""

from enum import IntEnum

class DrivingStrength(IntEnum):
    """SPIMaster Driving Strength

    Attributes:
        DS4MA:
        DS8MA:
        DS12MA:
        DS16MA:

    """
    DS4MA  = 0
    DS8MA  = 1
    DS12MA = 2
    DS16MA = 3

class Cpol(IntEnum):
    """SPIMaster Polarisation

    Attributes:
        IDLE_LOW: Idle low
        IDLE_HIGH: Idle high

    """
    IDLE_LOW  = 0
    IDLE_HIGH = 1

class Cpha(IntEnum):
    """SPIMaster Phase

    Attributes:
        CLK_LEADING: Leading phase
        CLK_TRAILING: Trailing phase

    """
    CLK_LEADING  = 0
    CLK_TRAILING = 1



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#
#cython: language_level=3

from __future__ import absolute_import
from libc.stdint cimport *
from .cftd2xx cimport *


cdef extern from "libft4222.h":
    ctypedef uint8_t  uint8
    ctypedef uint16_t uint16
    ctypedef uint32_t uint32
    ctypedef uint64_t uint64
    ctypedef int8_t   int8
    ctypedef int16_t  int16
    ctypedef int32_t  int32
    ctypedef int64_t  int64

    ctypedef enum FT4222_STATUS:
        FT4222_OK = 0
        FT4222_INVALID_HANDLE = 1
        FT4222_DEVICE_NOT_FOUND = 2
        FT4222_DEVICE_NOT_OPENED = 3
        FT4222_IO_ERROR = 4
        FT4222_INSUFFICIENT_RESOURCES = 5
        FT4222_INVALID_PARAMETER = 6
        FT4222_INVALID_BAUD_RATE = 7
        FT4222_DEVICE_NOT_OPENED_FOR_ERASE = 8
        FT4222_DEVICE_NOT_OPENED_FOR_WRITE = 9
        FT4222_FAILED_TO_WRITE_DEVICE = 10
        FT4222_EEPROM_READ_FAILED = 11
        FT4222_EEPROM_WRITE_FAILED = 12
        FT4222_EEPROM_ERASE_FAILED = 13
        FT4222_EEPROM_NOT_PRESENT = 14
        FT4222_EEPROM_NOT_PROGRAMMED = 15
        FT4222_INVALID_ARGS = 16
        FT4222_NOT_SUPPORTED = 17
        FT4222_OTHER_ERROR = 18
        FT4222_DEVICE_LIST_NOT_READY = 19

        FT4222_DEVICE_NOT_SUPPORTED = 1000        # FT_STATUS extending message
        FT4222_CLK_NOT_SUPPORTED = 1001
        FT4222_VENDER_CMD_NOT_SUPPORTED = 1002
        FT4222_IS_NOT_SPI_MODE = 1003
        FT4222_IS_NOT_I2C_MODE = 1004
        FT4222_IS_NOT_SPI_SINGLE_MODE = 1005
        FT4222_IS_NOT_SPI_MULTI_MODE = 1006
        FT4222_WRONG_I2C_ADDR = 1007
        FT4222_INVAILD_FUNCTION = 1008
        FT4222_INVALID_POINTER = 1009
        FT4222_EXCEEDED_MAX_TRANSFER_SIZE = 1010
        FT4222_FAILED_TO_READ_DEVICE = 1011
        FT4222_I2C_NOT_SUPPORTED_IN_THIS_MODE = 1012
        FT4222_GPIO_NOT_SUPPORTED_IN_THIS_MODE = 1013
        FT4222_GPIO_EXCEEDED_MAX_PORTNUM = 1014
        FT4222_GPIO_WRITE_NOT_SUPPORTED = 1015
        FT4222_GPIO_PULLUP_INVALID_IN_INPUTMODE = 1016
        FT4222_GPIO_PULLDOWN_INVALID_IN_INPUTMODE = 1017
        FT4222_GPIO_OPENDRAIN_INVALID_IN_OUTPUTMODE = 1018
        FT4222_INTERRUPT_NOT_SUPPORTED = 1019
        FT4222_GPIO_INPUT_NOT_SUPPORTED = 1020
        FT4222_EVENT_NOT_SUPPORTED = 1021
        FT4222_FUN_NOT_SUPPORT = 1022

    ctypedef enum FT4222_ClockRate:
        SYS_CLK_60 = 0
        SYS_CLK_24 = 1
        SYS_CLK_48 = 2
        SYS_CLK_80 = 3

    ctypedef enum I2C_MasterFlag:
        NONE = 0x80
        START = 0x02
        Repeated_START = 0x03  # Repeated_START will not send master code in HS mode
        STOP  = 0x04
        START_AND_STOP = 0x06  # START condition followed by SEND and STOP condition

    ctypedef enum GPIO_Trigger:
        GPIO_TRIGGER_RISING         = 0x01
        GPIO_TRIGGER_FALLING        = 0x02
        GPIO_TRIGGER_LEVEL_HIGH     = 0x04
        GPIO_TRIGGER_LEVEL_LOW      = 0X08

    ctypedef enum GPIO_Output:
        GPIO_OUTPUT_LOW = 0
        GPIO_OUTPUT_HIGH = 1

    ctypedef enum GPIO_Port:
        GPIO_PORT0  = 0
        GPIO_PORT1  = 1
        GPIO_PORT2  = 2
        GPIO_PORT3  = 3

    ctypedef enum GPIO_Dir:
        GPIO_OUTPUT = 0
        GPIO_INPUT  = 1

    ctypedef struct FT4222_Version:
        DWORD chipVersion
        DWORD dllVersion

    ctypedef enum FT4222_SPIMode:
        SPI_IO_NONE   = 0
        SPI_IO_SINGLE = 1
        SPI_IO_DUAL   = 2
        SPI_IO_QUAD   = 4

    ctypedef enum FT4222_SPIClock:
        CLK_NONE    = 0
        CLK_DIV_2   = 1   # 1/2   System Clock
        CLK_DIV_4   = 2   # 1/4   System Clock
        CLK_DIV_8   = 3   # 1/8   System Clock
        CLK_DIV_16  = 4   # 1/16  System Clock
        CLK_DIV_32  = 5   # 1/32  System Clock
        CLK_DIV_64  = 6   # 1/64  System Clock
        CLK_DIV_128 = 7   # 1/128 System Clock
        CLK_DIV_256 = 8   # 1/256 System Clock
        CLK_DIV_512 = 9   # 1/512 System Clock

    ctypedef enum FT4222_SPICPOL:
        CLK_IDLE_LOW  = 0
        CLK_IDLE_HIGH = 1

    ctypedef enum FT4222_SPICPHA:
        CLK_LEADING  = 0
        CLK_TRAILING = 1

    ctypedef enum SPI_DrivingStrength:
        DS_4MA  = 0
        DS_8MA  = 1
        DS_12MA = 2
        DS_16MA = 3

    ctypedef enum SPI_SlaveProtocol:
        SPI_SLAVE_WITH_PROTOCOL = 0
        SPI_SLAVE_NO_PROTOCOL   = 1
        SPI_SLAVE_NO_ACK        = 2


    FT4222_STATUS FT4222_UnInitialize(FT_HANDLE ftHandle)
    FT4222_STATUS FT4222_SetClock(FT_HANDLE ftHandle, FT4222_ClockRate clk)
    FT4222_STATUS FT4222_GetClock(FT_HANDLE ftHandle, FT4222_ClockRate* clk)
    FT4222_STATUS FT4222_SetWakeUpInterrupt(FT_HANDLE ftHandle, BOOL enable)
    FT4222_STATUS FT4222_SetInterruptTrigger(FT_HANDLE ftHandle, GPIO_Trigger trigger)
    FT4222_STATUS FT4222_SetSuspendOut(FT_HANDLE ftHandle, BOOL enable)
    FT4222_STATUS FT4222_GetMaxTransferSize(FT_HANDLE ftHandle, uint16* pMaxSize)
    FT4222_STATUS FT4222_SetEventNotification(FT_HANDLE ftHandle, DWORD mask, PVOID param)
    FT4222_STATUS FT4222_GetVersion(FT_HANDLE ftHandle, FT4222_Version* pVersion)
    # FT4222 I2C Functions
    FT4222_STATUS FT4222_I2CMaster_Read(FT_HANDLE ftHandle, uint16 deviceAddress, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred)
    FT4222_STATUS FT4222_I2CMaster_Write(FT_HANDLE ftHandle, uint16 deviceAddress, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred)
    FT4222_STATUS FT4222_I2CMaster_Init(FT_HANDLE ftHandle, uint32 kbps)
    FT4222_STATUS FT4222_I2CMaster_ReadEx(FT_HANDLE ftHandle, uint16 deviceAddress, uint8 flag, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred)
    FT4222_STATUS FT4222_I2CMaster_WriteEx(FT_HANDLE ftHandle, uint16 deviceAddress, uint8 flag, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred)
    FT4222_STATUS FT4222_I2CMaster_Reset(FT_HANDLE ftHandle)
    FT4222_STATUS FT4222_I2CMaster_GetStatus(FT_HANDLE ftHandle, uint8 *controllerStatus)
    # FT4222 GPIO Functions
    FT4222_STATUS FT4222_GPIO_Init(FT_HANDLE ftHandle, GPIO_Dir gpioDir[4])
    FT4222_STATUS FT4222_GPIO_Read(FT_HANDLE ftHandle, GPIO_Port portNum, BOOL* value)
    FT4222_STATUS FT4222_GPIO_Write(FT_HANDLE ftHandle, GPIO_Port portNum, BOOL bValue)
    FT4222_STATUS FT4222_GPIO_SetInputTrigger(FT_HANDLE ftHandle, GPIO_Port portNum, GPIO_Trigger trigger)
    FT4222_STATUS FT4222_GPIO_GetTriggerStatus(FT_HANDLE ftHandle, GPIO_Port portNum, uint16* queueSize)
    FT4222_STATUS FT4222_GPIO_ReadTriggerQueue(FT_HANDLE ftHandle, GPIO_Port portNum, GPIO_Trigger* events, uint16 readSize, uint16* sizeofRead)
    # FT4222 SPI Functions
    FT4222_STATUS FT4222_SPI_Reset(FT_HANDLE ftHandle);
    FT4222_STATUS FT4222_SPI_ResetTransaction(FT_HANDLE ftHandle, uint8 spiIdx);
    FT4222_STATUS FT4222_SPI_SetDrivingStrength(FT_HANDLE ftHandle, SPI_DrivingStrength clkStrength, SPI_DrivingStrength ioStrength, SPI_DrivingStrength ssoStrength);
    FT4222_STATUS FT4222_SPIMaster_Init(FT_HANDLE ftHandle, FT4222_SPIMode  ioLine, FT4222_SPIClock clock, FT4222_SPICPOL  cpol, FT4222_SPICPHA  cpha, uint8 ssoMap);
    FT4222_STATUS FT4222_SPIMaster_SetLines(FT_HANDLE ftHandle, FT4222_SPIMode spiMode);
    FT4222_STATUS FT4222_SPIMaster_SingleRead(FT_HANDLE ftHandle, uint8* buffer, uint16 bufferSize, uint16* sizeOfRead, BOOL isEndTransaction);
    FT4222_STATUS FT4222_SPIMaster_SingleWrite(FT_HANDLE ftHandle, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred, BOOL isEndTransaction);
    FT4222_STATUS FT4222_SPIMaster_SingleReadWrite(FT_HANDLE ftHandle, uint8* readBuffer, uint8* writeBuffer, uint16 bufferSize, uint16* sizeTransferred, BOOL isEndTransaction);
    FT4222_STATUS FT4222_SPIMaster_MultiReadWrite(FT_HANDLE ftHandle, uint8* readBuffer, uint8* writeBuffer, uint8 singleWriteBytes, uint16 multiWriteBytes, uint16 multiReadBytes, uint32* sizeOfRead);
    # FT4222 SPI Slave
    FT4222_STATUS FT4222_SPISlave_Init(FT_HANDLE ftHandle);
    FT4222_STATUS FT4222_SPISlave_InitEx(FT_HANDLE ftHandle, SPI_SlaveProtocol protocolOpt);
    FT4222_STATUS FT4222_SPISlave_SetMode(FT_HANDLE ftHandle, FT4222_SPICPOL  cpol, FT4222_SPICPHA  cpha);
    FT4222_STATUS FT4222_SPISlave_GetRxStatus(FT_HANDLE ftHandle, uint16* pRxSize);
    FT4222_STATUS FT4222_SPISlave_Read(FT_HANDLE ftHandle, uint8* buffer, uint16 bufferSize, uint16* sizeOfRead);
    FT4222_STATUS FT4222_SPISlave_Write(FT_HANDLE ftHandle, uint8* buffer, uint16 bufferSize, uint16* sizeTransferred);
    FT4222_STATUS FT4222_SPISlave_RxQuickResponse(FT_HANDLE ftHandle, BOOL enable);



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""SPI Master

Definitions to configure the SPI Master interface.
"""

from enum import IntEnum

class Mode(IntEnum):
    """SPIMaster Mode

    Attributes:
        SPI_IO_NONE: No IOs
        SPI_IO_SINGLE: Single mode
        SPI_IO_DUAL: Dual mode
        SPI_IO_QUAD: Quad mode

    """
    NONE   = 0
    SINGLE = 1
    DUAL   = 2
    QUAD   = 4

class Clock(IntEnum):
    """SPIMaster Clock

    Attributes:
        NONE:
        DIV_2: 1/2 System Clock
        DIV_4: 1/4 System Clock
        DIV_8: 1/8 System Clock
        DIV_16: 1/16 System Clock
        DIV_32: 1/32 System Clock
        DIV_64: 1/64 System Clock
        DIV_128: 1/128 System Clock
        DIV_256: 1/256 System Clock
        DIV_512: 1/512 System Clock

    """
    NONE    = 0
    DIV_2   = 1
    DIV_4   = 2
    DIV_8   = 3
    DIV_16  = 4
    DIV_32  = 5
    DIV_64  = 6
    DIV_128 = 7
    DIV_256 = 8
    DIV_512 = 9



class SlaveSelect(IntEnum):
    """Slave Slave bitmap

    Attributes:
        SS0: Slave select 0
        SS1: Slave select 1
        SS2: Slave select 2
        SS3: Slave select 3

    """
    SS0 = 1<<0
    SS1 = 1<<1
    SS2 = 1<<2
    SS3 = 1<<3



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.build_py import build_py
from Cython.Build import cythonize
from sys import platform as os_name
from platform import system, machine, architecture
import shutil
import os
import re

def find_version():
    tld = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(tld, 'ft4222', '__init__.py')
    with open(filename) as f:
        text = f.read()
    match = re.search(r"^__version__ = \"(.*)\"$", text, re.MULTILINE)
    if not match:
        raise RuntimeError('cannot find version')
    return match.group(1)

if system() ==  "Linux":
    if machine() == 'x86_64':
        libdir = "linux/build-x86_64"
    elif machine() == 'i386':
        libdir = "linux/build-i386"
    elif machine() == 'armv7l':
        libdir = "linux/build-arm-v7"
    elif machine().startswith("arm"):
        if architecture()[0] == '64bit':
            libdir = "linux/build-arm-v8"
        else:
            libdir = "linux/build-arm-v6-hf"
    elif machine() == 'aarch64':
        libdir = "linux/build-arm-v8"
    else:
        raise Exception("Unsupported machine and or architecture")

    libs = ["ft4222"]
    incdirs = ["linux"]
    libdirs = [libdir]
    rlibdirs = ['$ORIGIN/.']
    libs_to_copy = ["libft4222.so"]
elif system() == "Darwin":
    libdir = "./osx"
    ft4222_dll = "libft4222.dylib"

    libs = ["ft4222"] #, "ftd2xx"]
    incdirs = ["osx"]
    libdirs = [libdir]
    rlibdirs = [] #'$ORIGIN/.']
    libs_to_copy = [ft4222_dll, "libftd2xx.dylib", "libboost_system.dylib"]
else:
    if architecture()[0] == '64bit':
        libdir = "win/amd64"
        libs = ["LibFT4222-64", "ftd2xx"]
        libs_to_copy = ["LibFT4222-64.dll", "ftd2xx.dll"]
    else:
        libdir = "win/i386"
        libs = ["LibFT4222", "ftd2xx"]
        libs_to_copy = ["LibFT4222.dll", "ftd2xx.dll"]

    incdirs = ["win"]
    libdirs = [libdir]
    rlibdirs = []

class mybuild(build_py):
    def run(self):
        build_py.run(self)
        print("running mybuild")
        for package, src_dir, build_dir, filenames in self.data_files:
            if package == 'ft4222':
                for lib in libs_to_copy:
                    print("copying {} -> {}".format(libdir + "/" + lib, "ft4222/"+ lib))
                    shutil.copyfile(libdir + "/" + lib, build_dir + "/" + lib)
                break


extensions = [
    Extension("ft4222.ft4222", ["ft4222/ft4222.pyx"],
        libraries=libs,
        include_dirs=incdirs,
        library_dirs=libdirs,
        runtime_library_dirs=rlibdirs,
    ),
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ft4222',
    version=find_version(),
    author='Bearsh',
    author_email='me@bearsh.org',
    url='https://gitlab.com/msrelectronics/python-ft4222',
    description='Python wrapper around libFT4222.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'License :: Other/Proprietary License',
        'Operating System :: Microsoft',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Communications',
    ],
    keywords='ftdi ft4222',
    packages=['ft4222', 'ft4222.I2CMaster', 'ft4222.GPIO', 'ft4222.SPI', 'ft4222.SPIMaster', 'ft4222.SPISlave'],
    package_data={
        'ft4222': ['py.typed', 'ft4222.pyi', '__init__.pyi'],
        'ft4222.I2CMaster': ['py.typed'],
        'ft4222.GPIO': ['py.typed'],
        'ft4222.SPI': ['py.typed'],
        'ft4222.SPIMaster': ['py.typed'],
        'ft4222.SPISlave': ['py.typed'],
    },
    ext_modules=cythonize(extensions),
    cmdclass={'build_py': mybuild},
)



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#

"""Control FTDI USB chips for I2C/SPI/GPIO.

Open a handle using one of the ft4222.openBy... functions and use the methods
on the object thus returned.
"""

from __future__ import absolute_import
from .ft4222 import *
from .GPIO import *
from .I2CMaster import *
from .SPI import *
from .SPIMaster import *
from .SPISlave import *

__version__ = "1.5.1"

__all__ = [
    'FT2XXDeviceError',
    'FT4222DeviceError',
    'SysClock',
    'createDeviceInfoList',
    'getDeviceInfoDetail',
    'openBySerial',
    'openByDescription',
    'openByLocation',
    'FT4222',
]



#  _____ _____ _____
# |_    |   __| __  |
# |_| | |__   |    -|
# |_|_|_|_____|__|__|
# MSR Electronics GmbH
# SPDX-License-Identifier: MIT
#
#cython: language_level=3

from __future__ import absolute_import
from ft4222.cftd2xx cimport *
from ft4222.clibft4222 cimport *
from cpython.array cimport array, resize
from libc.stdio cimport printf
from enum import IntEnum
from .GPIO import Dir, Trigger
from .I2CMaster import ControllerStatus
from weakref import finalize as weakref_finalize

IF UNAME_SYSNAME == "Windows":
    cdef extern from "<malloc.h>" nogil:
        void *alloca(size_t size)
ELSE:
    cdef extern from "<alloca.h>" nogil:
        void *alloca(size_t size)


__ftd2xx_msgs = ['OK', 'INVALID_HANDLE', 'DEVICE_NOT_FOUND', 'DEVICE_NOT_OPENED',
                 'IO_ERROR', 'INSUFFICIENT_RESOURCES', 'INVALID_PARAMETER',
                 'INVALID_BAUD_RATE', 'DEVICE_NOT_OPENED_FOR_ERASE',
                 'DEVICE_NOT_OPENED_FOR_WRITE', 'FAILED_TO_WRITE_DEVICE0',
                 'EEPROM_READ_FAILED', 'EEPROM_WRITE_FAILED', 'EEPROM_ERASE_FAILED',
                 'EEPROM_NOT_PRESENT', 'EEPROM_NOT_PROGRAMMED', 'INVALID_ARGS',
                 'NOT_SUPPORTED', 'OTHER_ERROR', 'DEVICE_LIST_NOT_READY']

__ftd4222_msgs = ['DEVICE_NOT_SUPPORTED', 'CLK_NOT_SUPPORTED','VENDER_CMD_NOT_SUPPORTED',
                  'IS_NOT_SPI_MODE', 'IS_NOT_I2C_MODE', 'IS_NOT_SPI_SINGLE_MODE',
                  'IS_NOT_SPI_MULTI_MODE', 'WRONG_I2C_ADDR', 'INVAILD_FUNCTION',
                  'INVALID_POINTER', 'EXCEEDED_MAX_TRANSFER_SIZE', 'FAILED_TO_READ_DEVICE',
                  'I2C_NOT_SUPPORTED_IN_THIS_MODE', 'GPIO_NOT_SUPPORTED_IN_THIS_MODE',
                  'GPIO_EXCEEDED_MAX_PORTNUM', 'GPIO_WRITE_NOT_SUPPORTED',
                  'GPIO_PULLUP_INVALID_IN_INPUTMODE', 'GPIO_PULLDOWN_INVALID_IN_INPUTMODE',
                  'GPIO_OPENDRAIN_INVALID_IN_OUTPUTMODE', 'INTERRUPT_NOT_SUPPORTED',
                  'GPIO_INPUT_NOT_SUPPORTED', 'EVENT_NOT_SUPPORTED', 'FUN_NOT_SUPPORT']


# Revision A chips report chipVersion as 0x42220100; revision B chips report
# 0x42220200; revision C chips report 0x42220300. Revision B chips require
# version 1.2 or later of LibFT4222, indicated by dllVersion being greater than
# 0x01020000; Revision C chips require version 1.3 or later of LibFT4222, indicated
# by dllVersion being greater than 0x01030000.
__chip_rev_map = { 0x42220100: "Rev. A", 0x42220200: "Rev. B", 0x42220300: "Rev. C" }
__chip_rev_min_lib = { 0x42220100: 0, 0x42220200: 0x01020000, 0x42220300: 0x01030000 }


DEF MAX_DESCRIPTION_SIZE = 256

class FT2XXDeviceError(Exception):
    """Exception class for status messages"""
    def __init__(self, msgnum):
        self.message = __ftd2xx_msgs[msgnum]

    def __str__(self):
        return self.message

class FT4222DeviceError(FT2XXDeviceError):
    """Exception class for status messages"""
    def __init__(self, msgnum):
        if msgnum >= FT4222_DEVICE_NOT_SUPPORTED:
            self.message = __ftd4222_msgs[msgnum - FT4222_DEVICE_NOT_SUPPORTED]
        else:
            super(FT4222DeviceError, self).__init__(msgnum)

    def __str__(self):
        return self.message

class SysClock(IntEnum):
    """Chip system clock

    Attributes:
        CLK_60: 60 MHz
        CLK_24: 24 MHz
        CLK_48: 48 MHz
        CLK_80: 80 MHz

    """
    CLK_60 = 0
    CLK_24 = 1
    CLK_48 = 2
    CLK_80 = 3

def createDeviceInfoList():
    """Create the internal device info list and return number of entries"""
    cdef DWORD nb
    status = FT_CreateDeviceInfoList(&nb)
    if status == FT_OK:
        return nb
    raise FT2XXDeviceError, status

def getDeviceInfoDetail(devnum=0, update=True):
    """Get an entry from the internal device info list. Set update to
    False to avoid a slow call to createDeviceInfoList."""
    cdef:
        DWORD f, t, i, l
        FT_HANDLE h
        char n[MAX_DESCRIPTION_SIZE]
        char d[MAX_DESCRIPTION_SIZE]
    # createDeviceInfoList is slow, only run if update is True
    if update: createDeviceInfoList()
    status = FT_GetDeviceInfoDetail(devnum, &f, &t, &i, &l, n, d, &h)
    if status == FT_OK:
        return {'index': devnum, 'flags': f, 'type': t,
                'id': i, 'location': l, 'serial': n,
                'description': d, 'handle': <size_t>h}
    raise FT2XXDeviceError, status

def openBySerial(serial):
    """Open a handle to a usb device by serial number"""
    cdef FT_HANDLE handle
    cdef char* cserial = serial
    status = FT_OpenEx(<PVOID>cserial, FT_OPEN_BY_SERIAL_NUMBER, &handle)
    if status == FT_OK:
        return FT4222(<uintptr_t>handle, update=False)
    raise FT2XXDeviceError, status

def openByDescription(desc):
    """Open a handle to a usb device by description

    Args:
        desc (bytes, str): Description of the device

    Returns:
        :obj:`FT4222`: Opened device

    Raises:
        FT2XXDeviceError: on error

    """
    if isinstance(desc, str):
        desc = desc.encode('utf-8')
    cdef FT_HANDLE handle
    cdef char* cdesc = desc
    status = FT_OpenEx(<PVOID>cdesc, FT_OPEN_BY_DESCRIPTION, &handle)
    if status == FT_OK:
        #printf("handle: %d\n", handle)
        return FT4222(<uintptr_t>handle, update=False)
    raise FT2XXDeviceError, status

def openByLocation(locId):
    """Open a handle to a usb device by location

    Args:
        locId (int): Location id

    Returns:
        :obj:`FT4222`: Opened device

    Raises:
        FT2XXDeviceError: on error

    """
    cdef FT_HANDLE handle
    status = FT_OpenEx(<PVOID><uintptr_t>locId, FT_OPEN_BY_LOCATION, &handle)
    if status == FT_OK:
        return FT4222(<uintptr_t>handle, update=False)
    raise FT2XXDeviceError, status


cdef class _closer:
    """Wrap a hid_device *ptr and a provide a way to call hid_close() on it.
    Used internally for weakref.finalize, which only accepts Python objects.
    """

    cdef FT_HANDLE _handle

    @staticmethod
    cdef wrap(FT_HANDLE h):
        cdef _closer closer = _closer()
        closer._handle = h
        return closer

    def close(self):
        FT4222_UnInitialize(self._handle)
        FT_Close(self._handle)


cdef class FT4222:
    cdef FT_HANDLE _handle
    cdef DWORD _chip_version
    cdef DWORD _dll_version
    cdef object __weakref__  # enable weak-reference support
    cdef object _close

    def __init__(self, handle, update=True):
        self._handle = <FT_HANDLE><uintptr_t>handle
        self._chip_version = 0
        self._dll_version = 0
        self._get_version()
        self._close = weakref_finalize(self, _closer.wrap(self._handle).close)

    def close(self):
        """Closes the device."""
        FT4222_UnInitialize(self._handle)
        status = FT_Close(self._handle)
        if status != FT4222_OK:
            raise FT4222DeviceError, status
        self._handle = NULL

    cdef _get_version(self):
        cdef FT4222_Version ver
        status = FT4222_GetVersion(self._handle, &ver)
        if status == FT4222_OK:
            self._chip_version = ver.chipVersion
            self._dll_version = ver.dllVersion

    @property
    def chipVersion(self) -> int:
        """Chip version as number"""
        return self._chip_version

    @property
    def libVersion(self) -> int:
        """Library version as number"""
        return self._dll_version

    @property
    def chipRevision(self) -> str:
        """The revision of the chip in human readable format"""
        try:
            return __chip_rev_map[self._chip_version]
        except KeyError:
            return "Rev. unknown"

    def __repr__(self):
        return "FT4222: chipVersion: 0x{:x} ({:s}), libVersion: 0x{:x}".format(self._chip_version, self.chipRevision, self._dll_version)

    def setTimeouts(self, read_timeout, write_timeout):
        """Set the read and write timeouts

        Args:
            read_timeout (int): Read timeout in milliseconds.
            write_timeout (int): Write timeout in milliseconds.

        Raises:
            FT2XXDeviceError: on error

        """
        status = FT_SetTimeouts(self._handle, read_timeout, write_timeout)
        if status != FT_OK:
            raise FT2XXDeviceError, status

    def setClock(self, clk):
        """Set the system clock

        Args:
            clk (:obj:`ft4222.SysClock`): Desired clock

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SetClock(self._handle, clk)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def getClock(self):
        """Get the system clock

        Returns:
            :obj:`ft4222.SysClock`: Clock

        Raises:
            FT4222DeviceError: on error

        """
        cdef FT4222_ClockRate clk
        status = FT4222_GetClock(self._handle, &clk)
        if status == FT4222_OK:
            return SysClock(clk)
        raise FT4222DeviceError, status

    def setSuspendOut(self, enable):
        """Enable or disable, suspend out, which will emit a signal when FT4222H enters suspend mode.

        Args:
            enable (bool): TRUE to enable suspend out and configure GPIO2 as an output pin for emitting a signal when suspended. FALSE to switch back to GPIO2

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SetSuspendOut(self._handle, enable)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def setWakeUpInterrupt(self, enable):
        """Enable or disable the wakeup/interrupt

        Args:
            enable (bool): True to configure GPIO3 as an input pin for wakeup/interrupt. FALSE to switch back to GPIO3.

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SetWakeUpInterrupt(self._handle, enable)
        if status != FT4222_OK:
            raise FT4222DeviceError, status


    def vendorCmdGet(self, req, bytesToRead):
        """Vendor get command"""
        cdef:
            array[uint8] buf = array('B', [])
        resize(buf, bytesToRead)
        status = FT_VendorCmdGet(self._handle, req, buf.data.as_uchars, bytesToRead)
        if status == FT_OK:
            return bytes(buf)
        raise FT4222DeviceError, status

    def vendorCmdSet(self, req, data):
        """Vendor set command"""
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 bytesSent
            uint8* cdata = data
        status = FT_VendorCmdSet(self._handle, req, cdata, len(data))
        if status != FT_OK:
            raise FT4222DeviceError, status


    def gpio_Init(self, *args, gpio0=Dir.INPUT, gpio1=Dir.INPUT, gpio2=Dir.INPUT, gpio3=Dir.INPUT):
        """Initialize the GPIO interface.

        Args:
            *args (:obj:`list` of :obj:`ft4222.GPIO.Dir`, optional): List containing a direction for each port.
            gpio0 (:obj:`ft4222.GPIO.Dir`, optional): Direction of gpio0
            gpio1 (:obj:`ft4222.GPIO.Dir`, optional): Direction of gpio1
            gpio2 (:obj:`ft4222.GPIO.Dir`, optional): Direction of gpio2
            gpio3 (:obj:`ft4222.GPIO.Dir`, optional): Direction of gpio3

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            GPIO_Dir ioDir[4]
        if len(args) > 0:
            for i in xrange(len(args)):
                ioDir[i] = args[i]
            for i in xrange(len(args), 4 - len(args)):
                ioDir[i] = GPIO_INPUT
        else:
            ioDir[0] = gpio0
            ioDir[1] = gpio1
            ioDir[2] = gpio2
            ioDir[3] = gpio3
        status = FT4222_GPIO_Init(self._handle, ioDir)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def gpio_Read(self, portNum):
        """Read value from selected GPIO

        Args:
            portNum (:obj:`ft4222.GPIO.Port`): GPIO Port number

        Returns:
            bool: True if high, False otherwise

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            BOOL value
        status = FT4222_GPIO_Read(self._handle, portNum, &value)
        if status == FT4222_OK:
            return value
        raise FT4222DeviceError, status

    def gpio_Write(self, portNum, value):
        """Write value to given GPIO

        Args:
            portNum (:obj:`ft4222.GPIO.Port`): GPIO Port number
            value (bool): True for high, False for low

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_GPIO_Write(self._handle, portNum, value)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def gpio_SetInputTrigger(self, portNum, trigger):
        """Set software trigger conditions on the specified GPIO pin.

        Args:
            portNum (:obj:`ft4222.GPIO.Port`): GPIO Port number
            trigger (:obj:`ft4222.GPIO.Trigger`): Combination of trigger conditions

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_GPIO_SetInputTrigger(self._handle, portNum, trigger)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def gpio_GetTriggerStatus(self, portNum):
        """Get the size of trigger event queue.

        Args:
            portNum (:obj:`ft4222.GPIO.Port`): GPIO Port number

        Returns:
            int: Size of the trigger queue

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            uint16 queueSize
        status = FT4222_GPIO_GetTriggerStatus(self._handle, portNum, &queueSize)
        if status == FT4222_OK:
            return queueSize
        raise FT4222DeviceError, status

    def gpio_ReadTriggerQueue(self, portNum, readSize=None):
        """Get events recorded in the trigger event queue.

        Args:
            portNum (:obj:`ft4222.GPIO.Port`): GPIO Port number
            readSize (:obj:`int`, optional): Size of the queue, :obj:`gpio_GetTriggerStatus()` gets called if omitted

        Returns:
            :obj:`list` of :obj:`ft4222.GPIO.Trigger`: Trigger queue as list

        Raises:
            FT4222DeviceError: on error

        """
        if readSize == None:
            self.gpio_GetTriggerStatus(portNum)
        cdef:
            GPIO_Trigger *events = <GPIO_Trigger*>alloca(portNum * sizeof(GPIO_Trigger))
            uint16 sizeRead
        status = FT4222_GPIO_ReadTriggerQueue(self._handle, portNum, events, readSize, &sizeRead)
        if status == FT4222_OK:
            res = []
            for i in xrange(readSize):
                res.append(Trigger(events[i]))
            return res
        raise FT4222DeviceError, status


    def i2cMaster_Init(self, kbps=100):
        """Initialize the FT4222H as an I2C master with the requested I2C speed.

        Args:
            kbps (int): Speed in kb/s

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_I2CMaster_Init(self._handle, kbps)
        if status != FT4222_OK:
            raise FT4222DeviceError, status
        # current version (v1.3) of ftdi's lib can only handle clock rates down to 60kHz
        # although the chip can handle clock rates down to ~23.6kHz
        # there's a undocumented ways to achieve this
        if kbps < 60:
            #             Operating Clock Freq
            # SCL Freq = -----------------------  | M = 6 or 8; N = 1, 2, 3, …, 127
            #                 M*(N+1)
            #
            #       Operating Clock Freq             24MHz
            # N =  ---------------------- - 1   => ---------- - 1
            #            M * SCL Freq               8 * kbps
            #
            n = max(min(int(round(24000000.0 / (8 * 1000 * kbps) - 1)), 127), 1)
            self.setClock(SysClock.CLK_24)
            self.vendorCmdSet(0x52, n)

    def i2cMaster_Read(self, addr, bytesToRead):
        """Read data from the specified I2C slave device with START and STOP conditions.

        Args:
            addr (int): I2C slave address
            bytesToRead (int): Number of bytes to read from slave

        Returns:
            bytes: Bytes read from slave

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            array[uint8] buf = array('B', [])
            uint16 bytesRead
        resize(buf, bytesToRead)
        status = FT4222_I2CMaster_Read(self._handle, addr, buf.data.as_uchars, bytesToRead, &bytesRead)
        resize(buf, bytesRead)
        if status == FT4222_OK:
            return bytes(buf)
        raise FT4222DeviceError, status

    def i2cMaster_Write(self, addr, data):
        """Write data to the specified I2C slave device with START and STOP conditions.

        Args:
            addr (int): I2C slave address
            data (int, bytes, bytearray): Data to write to slave

        Returns:
            int: Bytes sent to slave

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 bytesSent
            uint8* cdata = data
        status = FT4222_I2CMaster_Write(self._handle, addr, cdata, len(data), &bytesSent)
        if status == FT4222_OK:
            return bytesSent
        raise FT4222DeviceError, status

    def i2cMaster_ReadEx(self, addr, flag, bytesToRead):
        """Read data from the specified I2C slave device with the specified I2C condition.

        Args:
            addr (int): I2C slave address
            flag (:obj:`ft4222.I2CMaster.Flag`): Flag to control start- and stopbit generation
            bytesToRead (int): Number of bytes to read from slave

        Returns:
            bytes: Bytes read from slave

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            array[uint8] buf = array('B', [])
            uint16 bytesRead
        resize(buf, bytesToRead)
        status = FT4222_I2CMaster_ReadEx(self._handle, addr, flag, buf.data.as_uchars, bytesToRead, &bytesRead)
        resize(buf, bytesRead)
        if status == FT4222_OK:
            return bytes(buf)
        raise FT4222DeviceError, status

    def i2cMaster_WriteEx(self, addr, flag, data):
        """Write data to the specified I2C slave device with the specified I2C condition.

        Args:
            addr (int): I2C slave address
            flag (:obj:`ft4222.I2CMaster.Flag`): Flag to control start- and stopbit generation
            data (int, bytes, bytearray): Data to write to slave

        Returns:
            int: Bytes sent to slave

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 bytesSent
            uint8* cdata = data
        status = FT4222_I2CMaster_WriteEx(self._handle, addr, flag, cdata, len(data), &bytesSent)
        if status == FT4222_OK:
            return bytesSent
        raise FT4222DeviceError, status

    def i2cMaster_Reset(self):
        """Reset the I2C master device.

        If the I2C bus encounters errors or works abnormally, this function will reset the I2C device.
        It is not necessary to call I2CMaster_Init again after calling this reset function.

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_I2CMaster_Reset(self._handle)
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def i2cMaster_GetStatus(self):
        """Read the status of the I2C master controller.

        This can be used to poll a slave until its write-cycle is complete.

        Returns:
            :obj:`ft4222.I2CMaster.ControllerStatus`: Controller status

        Raises:
            FT4222DeviceError: on error

        """
        cdef uint8 cs
        status = FT4222_I2CMaster_GetStatus(self._handle, &cs)
        if status == FT4222_OK:
            return ControllerStatus(cs)
        raise FT4222DeviceError, status


    def spi_Reset(self):
        """Reset the SPI master or slave device

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPI_Reset(self._handle);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spi_ResetTransaction(self, spiIdx):
        """Reset the SPI transaction

        Args:
            spiIdx (int): The index of the SPI transaction, which ranges from 0~3 depending on the mode of the chip.

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPI_ResetTransaction(self._handle, spiIdx);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spi_SetDrivingStrength(self, clkStrength, ioStrength, ssoStrength):
        """Reset the SPI master or slave device.

        Args:
            clkStrength (:obj:`ft4222.SPI.DrivingStrength`): Driving strength clock pin (master only)
            ioStrength (:obj:`ft4222.SPI.DrivingStrength`): Driving strength io pin
            ssoStrength (:obj:`ft4222.SPI.DrivingStrength`): Driving strength sso pin (master only)

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPI_SetDrivingStrength(self._handle, clkStrength, ioStrength, ssoStrength);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiMaster_Init(self, mode, clock, cpol, cpha, ssoMap):
        """Initialize as an SPI master under all modes.

        Args:
            mode (:obj:`ft4222.SPIMaster.Mode`): SPI transmission lines / mode
            clock (:obj:`ft4222.SPIMaster.Clock`): Clock divider
            cpol (:obj:`ft4222.SPI.Cpol`): Clock polarity
            cpha (:obj:`ft4222.SPI.Cpha`): Clock phase
            ssoMap (:obj:`ft4222.SPIMaster.SlaveSelect`): Slave selection output pins

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPIMaster_Init(self._handle, mode, clock, cpol, cpha, ssoMap);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiMaster_SetLines(self, mode):
        """Switch the FT4222H SPI master to single, dual, or quad mode.

        This overrides the mode passed to FT4222_SPIMaster_init. This might be needed if a
        device accepts commands in single mode but data transfer is to use dual or quad mode.

        Args:
            mode (:obj:`ft4222.SPIMaster.Mode`): SPI transmission lines / mode

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPIMaster_SetLines(self._handle, mode);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiMaster_SingleRead(self, bytesToRead, isEndTransaction):
        """Read data from a SPI slave in single mode

        Args:
            bytesToRead (int): Number of bytes to read
            isEndTransaction (bool): If True the slave select pin will be raised at the end

        Returns:
            bytes: Bytes read from slave

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            array[uint8] buf = array('B', [])
            uint16 bytesRead
        resize(buf, bytesToRead)
        status = FT4222_SPIMaster_SingleRead(self._handle, buf.data.as_uchars, bytesToRead, &bytesRead, isEndTransaction)
        if status == FT4222_OK:
            resize(buf, bytesRead)
            return bytes(buf)
        raise FT4222DeviceError, status

    def spiMaster_SingleWrite(self, data, isEndTransaction):
        """Write data to a SPI slave in single mode

        Args:
            data (bytes, bytearray, int): Data to write to slave
            isEndTransaction (bool): If True the slave select pin will be raised at the end

        Returns:
            int: Bytes sent to slave

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 bytesSent
            uint8* cdata = data
        status = FT4222_SPIMaster_SingleWrite(self._handle, cdata, len(data), &bytesSent, isEndTransaction);
        if status == FT4222_OK:
            return bytesSent
        raise FT4222DeviceError, status

    def spiMaster_SingleReadWrite(self, data, isEndTransaction):
        """Write and read data to and from a SPI slave in single mode

        Args:
            data (bytes, bytearray, int): Data to write to slave
            isEndTransaction (bool): If True the slave select pin will be raised at the end

        Returns:
            bytes: Bytes read from slave

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 sizeTransferred
            uint8* cdata = data
            array[uint8] buf = array('B', [])
        resize(buf, len(data))
        status = FT4222_SPIMaster_SingleReadWrite(self._handle, buf.data.as_uchars, cdata, len(data), &sizeTransferred, isEndTransaction);
        if status == FT4222_OK:
            resize(buf, sizeTransferred)
            return bytes(buf)
        raise FT4222DeviceError, status

    def spiMaster_MultiReadWrite(self, singleWrite, multiWrite, bytesToRead):
        """Write and read data to and from a SPI slave in dual- or quad-mode (multi-mode).

        Args:
            singleWrite (bytes, bytearray, int): Data to write to slave in signle-line mode (max. 15 bytes)
            multiWrite (bytes, bytearray, int): Data to write to slave in multi-line mode (max. 65535 bytes)
            bytesToRead (int):  Number of bytes to read on multi-line (max. 65535 bytes)

        Returns:
            bytes: Bytes read from slave in multi-line mode

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(singleWrite, int):
            data = bytes([singleWrite])
        elif not isinstance(singleWrite, (bytes, bytearray)):
            raise TypeError("the singleWrite argument must be of type 'int', 'bytes' or 'bytearray'")
        if isinstance(multiWrite, int):
            data = bytes([multiWrite])
        elif not isinstance(multiWrite, (bytes, bytearray)):
            raise TypeError("the multiWrite argument must be of type 'int', 'bytes' or 'bytearray'")
        write = singleWrite + multiWrite
        cdef:
            uint8* cdata = write
            array[uint8] buf = array('B', [])
            uint32 bytesRead
        resize(buf, bytesToRead)
        status = FT4222_SPIMaster_MultiReadWrite(self._handle, buf.data.as_uchars, cdata, len(singleWrite), len(multiWrite), bytesToRead, &bytesRead);
        if status == FT4222_OK:
            resize(buf, bytesRead)
            return bytes(buf)
        raise FT4222DeviceError, status

    def spiMaster_EndTransaction(self):
        """End the current SPI transaction.

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            DWORD bytesSent;
        status = FT_Write(self._handle, <unsigned char*>NULL, 0, &bytesSent);
        if status == FT_OK:
            return
        raise FT4222DeviceError, status

    def spiSlave_Init(self):
        """Initialize the FT4222H as an SPI slave. Default SPI_SlaveProtocol is SPI_SLAVE_WITH_PROTOCOL.

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPISlave_Init(self._handle);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiSlave_InitEx(self, proto):
        """Initialize as an SPI slave under all modes.

        Args:
            mode (:obj:`ft4222.SPISlave.Protocol`): SPI slave protocol

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPISlave_InitEx(self._handle,proto);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiSlave_Read(self, bytesToRead):
        """Read data from the receive queue of the SPI slave device.

        Args:
            bytesToRead (int): Number of bytes to read

        Returns:
            bytes: Bytes read from slave

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            array[uint8] buf = array('B', [])
            uint16 sizeRead
        resize(buf, bytesToRead)

        status = FT4222_SPISlave_Read(self._handle, buf.data.as_uchars, bytesToRead, &sizeRead);
        if status == FT4222_OK:
            resize(buf, sizeRead)
            return bytes(buf)
        raise FT4222DeviceError, status

    def spiSlave_SetMode(self, cpol, cpha):
        """Set SPI slave cpol and cpha. The Default value of cpol is (:obj:`ft4222.SPI.Cpol.CLK_IDLE_LOW`) , default value of cpha is (:obj:`ft4222.SPI.Cpol.CLK_LEADING`)

        Args:
            cpol (:obj:`ft4222.SPI.Cpol`): Clock polarity
            cpha (:obj:`ft4222.SPI.Cpha`): Clock phase

        Raises:
            FT4222DeviceError: on error

        """
        status = FT4222_SPISlave_SetMode(self._handle, cpol, cpha);
        if status != FT4222_OK:
            raise FT4222DeviceError, status

    def spiSlave_GetRxStatus(self):
        """Get number of bytes in the receive queue.

        Returns:
            pRxSize (uint16): Number of bytes in the receive queue

        Raises:
            FT4222DeviceError: on error

        """
        cdef:
            uint16 pRxSize

        status = FT4222_SPISlave_GetRxStatus(self._handle, &pRxSize);

        if status == FT4222_OK:
            return pRxSize
        raise FT4222DeviceError, status

    def spiSlave_Write(self, data):
        """Write data to the transmit queue of the SPI slave device.

        Args:
            data (bytes, bytearray, int): Data to write to slave

        Returns:
            sizeTransferred (uint16): Number of bytes written to the device.

        Raises:
            FT4222DeviceError: on error

        """
        if isinstance(data, int):
            data = bytes([data])
        elif not isinstance(data, (bytes, bytearray)):
            raise TypeError("the data argument must be of type 'int', 'bytes' or 'bytearray'")
        cdef:
            uint16 sizeTransferred
            uint8* cdata = data

        status = FT4222_SPISlave_Write(self._handle, cdata, len(data), &sizeTransferred);
        if status == FT4222_OK:
            return sizeTransferred
        raise FT4222DeviceError, status
